/*
 *    X2  IRC Operator and Channel services
 *    by Rubin (rubin@afternet.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "misc.h"
#include "user.h"
#include "sock.h"
#include "chan.h"
#include "send.h"
#include "match.h"
#include "server.h"

char globsign;

void SetTopics(int period)
{
  int i;
  ChannelListType ChanPtr;

  for (i = 0; i < CHANSEGSIZE; i++)
  {
    ChanPtr = Channels->Table[i];
    while (ChanPtr)
    {
      if ((ChanPtr->Regged && !ChanPtr->Suspended && ChanPtr->settings[INCHAN] == 1) && ChanPtr->settings[TOPICFREQ] == period)
      {
        if(strcmp(ChanPtr->Topic, "*"))
          send_topic(ChanPtr, ChanPtr->Topic);
      }
      ChanPtr = ChanPtr->Next;
    }
  }
}


ChannelListType GetChannel(char *Channel)
{

  int Pos;
  unsigned long Key;
  ChannelListType ChanPtr;

  if(!Channel || !*Channel)
      return(NULL);

  Key = Hash(&Channel[1]);
  Pos = Key % CHANSEGSIZE;

  ChanPtr = Channels->Table[Pos];
  while (ChanPtr)
  {
    if (!strcasecmp(ChanPtr->Name, Channel))
    {
      return (ChanPtr);
    }
    ChanPtr = ChanPtr->Next;
  }
  return (NULL);
}

/* If there's a mask that matches hostmask, return the numeric 
 * value of the level of that user. */
int IsProt(ChannelListType ChanPtr, char *hostmask)
{
   HandleListType hptr;
   MaskListType mptr;

   if(ChanPtr == NULL)
      return(0);
   for(hptr = ChanPtr->Handles;hptr;hptr = hptr->Next)
   {
       for(mptr = hptr->Masks;mptr;mptr = mptr->Next)
       {
            if(!matches(mptr->Mask, hostmask) || !matches(hostmask,mptr->Mask))
            {
                  return(userlevelbychar(hptr->Level));
            }
       }
   }
   return(0);
}

int DelLamer(ChannelListType cptr, char* Mask)
{
  LamerListType lptr, lptr_Next, lptr_prev = NULL;

  for(lptr = cptr->Lamers;lptr;lptr = lptr_Next)
  {
     lptr_Next = lptr->Next;
     if(!strcasecmp(lptr->Mask, Mask))
     {
        if(lptr_prev == NULL)
        {
           cptr->Lamers = lptr->Next;
        }
        else
        {
           lptr_prev->Next = lptr->Next;
        }
        --LamerCount;
        free(lptr->Mask);
        free(lptr->Reason);
        free(lptr);
        return(TRUE);
     }
     lptr_prev = lptr;
  }
  return(FALSE);
}

void AddLamer(ChannelListType cptr, time_t BanTime, time_t LastSeen, char* Address, char *Reason)
{
  LamerListType lptr;
  /* Add to the channel's lamer list. */

  lptr = malloc(sizeof(struct LamerStruct));
  lptr->Mask = malloc(strlen(Address)+1);
  strcpy(lptr->Mask, Address);

  lptr->Reason = malloc(strlen(Reason)+1);
  strcpy(lptr->Reason, Reason);

  lptr->Time = time(NULL);
  lptr->LastSeen = 0;

  lptr->Next = cptr->Lamers;
  cptr->Lamers = lptr;

  ++LamerCount;

  WriteChanOpFile(cptr);
}

LamerListType GetLamer(ChannelListType cptr, char *mask)
{
  LamerListType lptr;

  for(lptr = cptr->Lamers;lptr;lptr = lptr->Next)
  {
     if(!strcasecmp(lptr->Mask, mask))
     {
        return(lptr);
     }
  }
  return(NULL);
}

LamerListType IsChanLamer(ChannelListType cptr, UserListType uptr)
{
  LamerListType lptr;
  char *nuh;

  if(uptr == NULL || cptr == NULL)
     return(NULL);

  nuh = (char *) malloc(strlen(uptr->Nick) + strlen(uptr->UserID) + strlen(uptr->Address) + 3);
  sprintf(nuh, "%s!%s@%s", uptr->Nick, uptr->UserID, uptr->Address);

  for(lptr = cptr->Lamers;lptr;lptr = lptr->Next)
  {
     if(!matches(lptr->Mask, nuh))
     {
         free(nuh);
         return(lptr);
     }
  }
  free(nuh);
  return (NULL);
}

int IsChanAOp(HandleListType hptr)
{
  if(hptr && hptr->AutoOp)
    return (1);
  else
    return (0);
}

void SetChanAOp(HandleListType hptr, int setting)
{
  hptr->AutoOp = setting;
  return;
}

void SetInfo(ChannelListType cptr, HandleListType hptr, char *Info)
{
  if (strlen(Info) >= MAXINFOLEN)
  {
      Info[MAXINFOLEN] = '\0';
  }
  free(hptr->Info);
  hptr->Info = (char *) malloc(strlen(Info) + 1);
  strcpy(hptr->Info, Info);
  WriteChanOpFile(cptr);
}

void SetLastSeen(ChannelListType cptr, HandleListType hptr, int action)
{
  if(hptr == NULL)
    return;
  hptr->LastSeen = time(NULL);
  if (action == JOIN)
    hptr->InChan = 1;
  if (action == PART)
    hptr->InChan = 0;
  WriteChanOpFile(cptr);
}

int isjoined(ChannelListType cptr, UserListType uptr)
{
   JoinedChansListType jptr;
   
   for(jptr = uptr->Chans;jptr;jptr = jptr->Next)
   {
       if(jptr->cptr == cptr)
          return(TRUE);
   }
   return(FALSE);
}

int IsInChan(HandleListType hptr)
{
   if (hptr->InChan > 0)
       return (TRUE);
   else
       return (FALSE);
}

void SendInfoLine(ChannelListType cptr, HandleListType hptr, UserListType uptr)
{
  if (hptr != NULL && strcmp(hptr->Info, "*"))
     send_privmsg(cptr->Name, "[%s] %s", uptr->Nick, hptr->Info);
}

char *GetInfo(HandleListType hptr)
{
  return (hptr->Info);
}

/* First we check if uptr has a login to the channel active. If so we return that hptr.
 * if not, we try to find a good (the best) matching handle for the person.
 */

HandleListType FindBestUserMatch(ChannelListType cptr, UserListType uptr)
{
  char Str[LINEBUFFSIZE];
  int weight, prevweight = 0, lvl, prevlvl = 0;
  HandleListType hptr;
  MaskListType mptr;
  HandleListType PrevHandlePtr=NULL;
  AuthHandleListType ahptr;

  if(cptr == NULL)
  { /* Maybe no channel given, or regchan... */
    return (NULL);
  }
  if (uptr == NULL)
  {
    return (NULL);
  }
  if(!cptr->Regged)
  {
     return(NULL); /* Not registered. */
  }
  /* Look for a login. */
  for(ahptr = uptr->AuthHandles; ahptr; ahptr = ahptr->Next)
  {
      if(ahptr->cptr == cptr)
      {
         return(ahptr->hptr);
      }
  }
  /* No login, so find a match here... */
  sprintf(Str, "%s!%s@%s", uptr->Nick, uptr->UserID, uptr->Address);
  for(hptr = cptr->Handles;hptr;hptr = hptr->Next)
  {
    for(mptr = hptr->Masks;mptr;mptr = mptr->Next)
    {
      if (!matches(mptr->Mask, Str))
      { /* If mask matches our user, */
         weight=calcmaskweight(mptr->Mask);
         lvl=userlevelbychar(hptr->Level);
         if ( (!PrevHandlePtr || (lvl > prevlvl) || (weight > prevweight && lvl >= prevlvl)))
         {
            PrevHandlePtr = hptr;
            prevweight = weight;
            prevlvl = lvl;
         }
      }
    }
  }
  return(PrevHandlePtr);
  return(NULL);
}

int PublicReplyAllowed(HandleListType hptr)
{
   if(userlevel(hptr) >= PE)
      return(TRUE);
   return(FALSE);
}

boolean IsChanOwner(HandleListType hptr)
{
   if(hptr && hptr->Level == 'W')
     return(TRUE);
   return(FALSE);
}

boolean IsChanCoOwner(HandleListType hptr)
{
   if(hptr && hptr->Level == 'C')
     return(TRUE);
   return(FALSE);
}

boolean IsChanManager(HandleListType hptr)
{
   if(hptr && hptr->Level == 'M')
     return(TRUE);
   return(FALSE);
}

boolean IsChanOp(HandleListType hptr)
{
   if(hptr && hptr->Level == 'O')
     return(TRUE);
   return(FALSE);
}

boolean IsChanPeon(HandleListType hptr)
{
   if(hptr && hptr->Level == 'P')
     return(TRUE);
   return(FALSE);
}

boolean IsDeop(char *Modes)
{
  char Operator = '+';
  int Len, i;

  Len = strlen(Modes);
  for (i = 0; i < Len; i++)
  {
    if (Modes[i] == '-')
      Operator = '-';
    if ((Modes[i] == 'o') && (Operator == '-'))
      return (true);
  }
  return (false);
}

boolean IsBan(char *Modes)
{
  char Operator = '+';
  int Len, i;

  Len = strlen(Modes);
  for (i = 0; i < Len; i++)
  {
    if (Modes[i] == '-')
      Operator = '-';
    if (Modes[i] == '+')
      Operator = '+';
    if ((Modes[i] == 'b') && (Operator == '+'))
      return (true);
  }
  return (false);
}

void ProcessKick(ChannelListType cptr, UserListType uptr, char *Kicked)
{
  UserListType kptr;
  if (strcasecmp(Kicked, IrcNick) == 0)
  {				/* If it's me who got kicked.. */
    send_part(cptr, "Kicked");  /* This is required by protocol */
    send_wallops("\002Ok, very funny.  Now, tell %s to stop kicking me from %s!\002", uptr->Nick, cptr);
    send_join(cptr);
    send_opchan(cptr);
  }
  if (cptr == NULL)
  {
    Debug(DBGWARNING, "ProcessKick(): NULL Channel");
    return;    /* chan not found??*/
  }    
  if(!(kptr = FindNumnick(Kicked, UsersU)))
  {
    Debug(DBGWARNING, "ProcessKick(): Cannot find target %s", Kicked);
    return;
  }
  if (cptr->settings[INCHAN] != 0 && !kickok(cptr, uptr, kptr) && !IsModeK(uptr))
    punish(cptr, uptr, kptr, 'K');

  RemoveNickFromChannel(cptr, kptr); /* NOTE: This can delete cptr
                                      * if that was the last user. */
}

void on_mode(ChannelListType cptr, UserListType uptr, char *rest)
{
  int did_change = FALSE;	/* If we have to change anything */
  char *mode;

  char *chanchars;		/* i.e. +oobli */
  char *params;
  char sign = '+';		/* +/-, i.e. +o, -b etc */

  char unminmode[MAXLEN];	/* used to undo modechanges */
  char unminparams[MAXLEN];
  char unplusmode[MAXLEN];	/* used to undo modechanges */
  char unplusparams[MAXLEN];

  /* these are here merely to make things clearer */
  char *numnick = NULL;
  char *banstring;
  char *key;
  char *limit;
  static time_t currenttime;

  currenttime = getthetime();

  strcpy(unminmode, "");	/* min-modes, i.e. -bi */
  strcpy(unminparams, "");	/* it's parameters */
  strcpy(unplusmode, "");	/* plus-modes, i.e. +oo */
  strcpy(unplusparams, "");	/* parameters */

  if (cptr == NULL)		/* if I'm not in that chan */
  {
    Debug(DBGWARNING, "on_mode(): Mode with NULL channel.");
    return;
  }
  if(!cptr->settings[INCHAN])
    return;
  mode = rest; 
  params = (char *) strchr(mode, ' ') + 1;
  chanchars = (char *) mysep(&mode, " ");
  /* if we hit a space, stop. */
  while (*chanchars && *chanchars!=' ')
  {
    switch (*chanchars)
    {
    case '+':
      sign = '+';
      globsign = '+';
      break;
    case '-':
      sign = '-';
      globsign = '-';
      break;
    case 'o':
    {
      UserListType optr;
      numnick = (char *) strtok(params, " ");
      params = (char *) strtok(NULL, "\0\n");
      if (!numnick)
	break;
      if (IsModeK(uptr))
	break;
/* server can do modes.. hmm.
      if (GetServer(uptr))
	break;
*/
      optr = FindNumnick(numnick, UsersU);
      if (sign == '+')
      {
	/* Do checks if we should remove the o */
	if (!opok(cptr, uptr, optr))
	{
	  strcat(unminmode, "o");
	  strcat(unminparams, numnick);
	  strcat(unminparams, " ");
	  did_change = TRUE;
	}
      }
      else
	/* sign == '-' */
      {
	/* Do checks if we should give back the o */
	if (!deopok(cptr, uptr, optr))
	{
	  strcat(unplusmode, "o");
	  strcat(unplusparams, numnick);
	  strcat(unplusparams, " ");
	  did_change = TRUE;
          punish(cptr, uptr, optr, 'o');
	}
      }
      break;
    }
    case 'b':
      banstring = (char *) strtok(params, " ");
      params = (char *) strtok(NULL, "\0\n");
      if (!banstring)
	break;
      if (sign == '+')
      {
	if (!IsModeK(uptr)) /* If its not a service doing it */
	  if (!banok(cptr, uptr, banstring))
	  {
	    strcat(unminmode, "b");
	    strcat(unminparams, banstring);
	    strcat(unminparams, " ");
	    did_change = TRUE;
//	    punish(cptr, uptr, banstring, 'b');
	    break;
	  }
	add_ban(cptr, banstring);
      }
      else
	/* sign == '-' */
      {
	delete_ban(cptr, banstring);
      }
      break;
    case 'v':
      numnick = (char *) strtok(params, " ");
      params = (char *) strtok(NULL, "\0\n");
      break;
    case 'p':
      if (modeok(cptr, uptr))
	break;
      if (sign == '+')
      {
        if (cptr->MinModes & CMODEp)
	{
	  strcat(unminmode, "p");
	  did_change = TRUE;
	}
      }
      else
	/* sign == '-' */
      {
	if (cptr->PlusModes & CMODEp)
	{
	  strcat(unplusmode, "p");
	  did_change = TRUE;
	}
      }
      break;
    case 's':
      if (modeok(cptr, uptr))
	break;
      if (sign == '+')
      {
	if (cptr->MinModes & CMODEs)
	{
	  strcat(unminmode, "s");
	  did_change = TRUE;
	}
      }
      else
	/* sign == '-' */
      {
	if (cptr->PlusModes & CMODEs)
	{
	  strcat(unplusmode, "s");
	  did_change = TRUE;
	}
      }
      break;
    case 'm':
      if (modeok(cptr, uptr))
	break;
      if (sign == '+')
      {
	if (cptr->MinModes & CMODEm)
	{
	  strcat(unminmode, "m");
	  did_change = TRUE;
	}
      }
      else
	/* sign == '-' */
      {
	if (cptr->PlusModes & CMODEm)
	{
	  strcat(unplusmode, "m");
	  did_change = TRUE;
	}
      }
      break;
    case 't':
      if (modeok(cptr, uptr))
	break;
      if (sign == '+')
      {
	if (cptr->MinModes & CMODEt)
	{
	  strcat(unminmode, "t");
	  did_change = TRUE;
	}
      }
      else
	/* sign == '-' */
      {
	if (cptr->PlusModes & CMODEt)
	{
	  strcat(unplusmode, "t");
	  did_change = TRUE;
	}
      }
      break;
    case 'i':
      if (modeok(cptr, uptr))
	break;
      if (sign == '+')
      {
	if (cptr->MinModes & CMODEi)
	{
	  strcat(unminmode, "i");
	  did_change = TRUE;
	}
      }
      else
	/* sign == '-' */
      {
	if (cptr->PlusModes & CMODEi)
	{
	  strcat(unplusmode, "i");
	  did_change = TRUE;
	}
      }
      break;
    case 'n':
      if (modeok(cptr, uptr))
	break;
      if (sign == '+')
      {
	if (cptr->MinModes & CMODEn)
	{
	  strcat(unminmode, "n");
	  did_change = TRUE;
	}
      }
      else
	/* sign == '-' */
      {
	if (cptr->PlusModes & CMODEn)
	{
	  strcat(unplusmode, "n");
	  did_change = TRUE;
	}
      }
      break;
    case 'k':
      key = (char *) strtok(params, " ");
      params = (char *) strtok(NULL, "\0\n");
      if(!key)
      {
         Debug(DBGWARNING, "NULL channel key in on_mode()");
         break;
      }
      if(sign == '+') 
          safecopy(cptr->CurrentKey, key, 25);
          
      if (modeok(cptr, uptr))
	break;
      if (sign == '+')
      {
	strncpy(cptr->CurrentKey, key, 25);
	if (cptr->MinModes & CMODEk)
	{
	  strcat(unminmode, "k");
	  strcat(unminparams, key);
	  strcat(unminparams, " ");
	  did_change = TRUE;
	}
      }
      else
	/* sign == '-' */
      {
	strcpy(cptr->CurrentKey, "");
	if (cptr->PlusModes & CMODEk)
	{
	  strcat(unplusmode, "k");
	  strcat(unplusparams, cptr->Key);
	  strcat(unplusparams, " ");
	  did_change = TRUE;
	}
      }
      break;
    case 'l':
      if (modeok(cptr, uptr))
	break;
      if (sign == '+')
      {
	limit = (char *) strtok(params, " ");
	params = (char *) strtok(NULL, "\0\n");
	if (cptr->MinModes & CMODEl)
	{
	  strcat(unminmode, "l");
	  did_change = TRUE;
	}
	else if ((cptr->PlusModes & CMODEl) && strcmp(limit, cptr->Limit))
	{
	  strcat(unplusmode, "l");
	  strcat(unplusparams, cptr->Limit);
	  strcat(unplusparams, " ");
	  did_change = TRUE;
	}
      }
      else /* sign == '-' */
      {
	if (cptr->PlusModes & CMODEl)
	{
	  strcat(unplusmode, "l");
/* -l does not take a parameter.
 *	  strcat(unplusparams, cptr->Limit);
 */	  strcat(unplusparams, " ");
	  did_change = TRUE;
	}
      }
      break;
    default:
      Debug(DBGINFO, "Unknown channel mode %c [%x] in %s", *chanchars, *chanchars, cptr->Name);
      break;
    }
    chanchars++;
  }
  /* restore unwanted modechanges */
  if (did_change)
  {
    send_chanmode(cptr, "-%s+%s %s %s", unminmode, unplusmode, unminparams, unplusparams);
  }
  return;
}

void ConvertModeString(char *ModeString, int *PlusModes, int *MinModes, char **Key, char **limit)
{
  int *modeptr, *optr;
  char *modes;
  char *tmpstr;

  if(!ModeString || !*ModeString)
     return;
  modes = (char *) mysep(&ModeString, " ");
  modeptr = PlusModes;
  optr = MinModes;
  *PlusModes = 0;
  *MinModes = 0;
  if (*Key)
  {
    free(*Key);
    *Key = NULL;
  }
  if (*limit)
  {
    free(*limit);
    *limit = NULL;
  }
  while (*modes)
  {
    switch (*modes)
    {
    case '+':
      modeptr = PlusModes;
      optr = MinModes;
      break;
    case '-':
      modeptr = MinModes;
      optr = PlusModes;
      break;
    case 'n':
      *modeptr = *modeptr | CMODEn;
      *optr = *optr & ~CMODEn;
      break;
    case 't':
      *modeptr = *modeptr | CMODEt;
      *optr = *optr & ~CMODEt;
      break;
    case 's':
      *modeptr = *modeptr | CMODEs;
      *optr = *optr & ~CMODEs;
      break;
    case 'p':
      *modeptr = *modeptr | CMODEp;
      *optr = *optr & ~CMODEp;
      break;
    case 'i':
      *modeptr = *modeptr | CMODEi;
      *optr = *optr & ~CMODEi;
      break;
    case 'm':
      *modeptr = *modeptr | CMODEm;
      *optr = *optr & ~CMODEm;
      break;
    case 'k':
      if (modeptr == MinModes)
      {
	*modeptr = *modeptr | CMODEk;
	*optr = *optr & ~CMODEk;
	break;
      }
      if((tmpstr = (char *) mysep(&ModeString, " ")))
      {
	if (KeyOk(tmpstr))
	{
	  if (*Key)
	    break;		/* already got first one.. */
	  *Key = malloc(strlen(tmpstr) + 2);
	  if (!(*Key))
	    return;		/* out of memory! */
	  strcpy(*Key, tmpstr);
	  *modeptr = *modeptr | CMODEk;
	}
      }
      break;
    case 'l':
      if (modeptr == MinModes)
      {
	*modeptr = *modeptr | CMODEl;
	*optr = *optr & ~CMODEl;
	break;
      }
      if((tmpstr = (char *) mysep(&ModeString, " ")))
      {
	if (IsAllDigits(tmpstr) && (atol(tmpstr) != 0))
	{
	  if (*limit)
	    break;		/* already got firstone */
	  *limit = malloc(strlen(tmpstr) + 2);
	  if (!(*limit))
	    return;		/* out of memory! */
	  strcpy(*limit, tmpstr);
	  *modeptr = *modeptr | CMODEl;
	}
      }
      break;
    default:
      break;
    }
    modes++;
  }
}

void ProcessJoin(ChannelListType ChanNode, UserListType UserNode, char* channel, char type)
{		/* on_join */
  /* note anything that calls strtok cannot be used
   * in this function or it will break the calling
   * function. use mysep(). 
   * 
   * type is 'C' for CREATE, 'B' for BURST, and 'J' for JOIN
   */

  ChannelUsersListType NewCUser;
  JoinedChansListType jptr;
  int gaveops=0;
  time_t seentime;
  HandleListType hptr;
  LamerListType lptr;

  if (!UserNode)
    return;
  if(!(ChanNode))
  {
     if (!AddChan(channel, FALSE, 0, NULL, NULL, NULL, NULL, NULL,NULL, NetTime(), 0))
     {
          Debug(DBGWARNING, "Error adding channel %s in ProcessJoin()", channel);
          return;
     }
     ChanNode = GetChannel(channel);
  }
  if (!(ChanNode))
  { 
     Debug(DBGWARNING, "ProcessJoin(): Added channel unfindable!!");
     return;
  }
  Debug(DBGINFO, "Processing Join: %s %s %s", 
                  UserNode->Numnick, (type == 'B')?"burstjoined":(type=='C')?"created":"joined", ChanNode->Name);
  ChanNode->NumOfNicks++; /* Incriment ppl in chan counter */

  NewCUser = malloc(sizeof(struct ChanUserStruct));
  NewCUser->User = UserNode;
  NewCUser->Next = ChanNode->Joined;
  ChanNode->Joined = NewCUser;

  jptr = malloc(sizeof(struct JoinedChansStruct));
  jptr->cptr = ChanNode;
  jptr->Next = UserNode->Chans;
  UserNode->Chans = jptr;
/** Done **/

  if(ChanNode->Regged == FALSE)
  {  /* This channel is not regged */
     return;
  }
  if(!ChanNode->settings[INCHAN])                                                                   
    return; /* Channel not active */
  if(type != 'B')
  {
     if(strcmp(ChanNode->URL, "*"))
     {
         send_numeric(328, UserNode->Numnick, "%s :%s", ChanNode->Name, ChanNode->URL);
     }
  }
  if (ChanNode->settings[AUTOOP] == 3)
  {
    send_op(ChanNode, UserNode);
    gaveops++;
  }

  if((hptr = FindBestUserMatch(ChanNode, UserNode)))
  {
    seentime = hptr->LastSeen;
  }
  else 
    seentime = 0;
  if (seentime != 0)
  {                                        /* CO = CO OWNER */
    if ((seentime < NetMessage.Timestamp) && userlevel(hptr) >= CO)
      SendNetMessage(UserNode);
  }
  SetLastSeen(ChanNode, hptr, JOIN);
  /* If there password is '*' ask them to change it.*/

  if(hptr && !strcmp(crypt("*","X2"), hptr->Pass))
  {
      send_to_user(UserNode, "This version of X2 requires that all users have passwords set.");
      send_to_user(UserNode, "Please set your password. T ype: /msg %s %s PASS * yournewpass", IrcNick, ChanNode->Name);
  }

/* If they are (co)owner, mark that one has joined. */
  if(IsChanOwner(hptr) || IsChanCoOwner(hptr))
  {
     ChanNode->OwnersOnline++;   
     if(ChanNode->OwnersOnline > 0)
        ChanNode->LastAccess = (unsigned long) 0;
     WriteChanOpFile(ChanNode);
  }

  /* If they do not have an email address set, nag them. */
  if(IsChanOwner(hptr))
  {
     if(ChanNode->Email[0] == '*')
     { 
         send_to_user(UserNode, "This version of X2 requires a valid email address be set for each channel.");
         send_to_user(UserNode, "That way, if you forget a password, X2 can automatically email you to recover access.");
         send_to_user(UserNode, "Please .set email your@email or /msg %s set %s email your@emailaddy.", IrcNick, ChanNode->Name);
         send_to_user(UserNode, "Your email address will be used ONLY for identification or Channel service related email notices. We won't let spammers get ahold of it (Indeed, only you and our IRC Operators will even be able to see it.");
     }
  }

/* Figure out if we should op them.. */
  if(IsChanAOp(hptr))
  {
    switch(userlevel(hptr))
    {
      case PE: /* Peon */
            if(ChanNode->settings[VOICE] == 1)
                send_voice(ChanNode, UserNode);
            break;
      case OP: /* OP */
           if(ChanNode->settings[AUTOOP] > 1)
           {
                send_op(ChanNode, UserNode);
                gaveops++;
           }
           break;
      /* These 3 share the same op rank. 0=dont op, anything else, op them. */
      case MA: /* Manager */
      case CO: /* CoOwner */
      case OW: /* Owner */
           if(ChanNode->settings[AUTOOP] > 0)
           {
                send_op(ChanNode, UserNode);
                gaveops++;
           }
           break;
      default: /* The level is 0 or unknown? */
           if(ChanNode->settings[AUTOOP] == 3) /* Everyone */
           {
                send_op(ChanNode, UserNode);
                gaveops++;
           }
    }
  }
/* Send their info line, if it wasnt a BURST (net re-join) */
    if(hptr != NULL && type != 'B' && ChanNode->settings[USERINFO] == 1)
      SendInfoLine(ChanNode, hptr, UserNode);

/* If they aren't a user, but they do match a ban... */
  if(hptr == NULL && (lptr = IsChanLamer(ChanNode, UserNode)))
  {
    if(IsModeK(UserNode)) /* Service bot, or me, dont harm it */
      return;
    deopban(ChanNode, UserNode, lptr->Mask);
    send_kick(ChanNode, UserNode, "%s", lptr->Reason);
    return; /* So as not to voice them too! */
  }
/* If we didn't op them already, and X2 is set to voice "everyone", 
 * then give them voice! */
  if(ChanNode->settings[VOICE] == 2)
  {
     if(!gaveops)
       send_voice(ChanNode, UserNode);
  }
}

void RemoveNickFromChannel(ChannelListType cptr, UserListType uptr)
{
  ChannelUsersListType ChannelUser, TempCUser, PrevUser;
  int DeletedChanUsers = 0;
  HandleListType hptr;
  JoinedChansListType jptr, jptr_prev = NULL;

  if (!cptr)
  {
    Debug(DBGWARNING, "RemoveNickFromChannel(): NULL channel.");
    return;
  }
  if(!uptr)
  {
    Debug(DBGWARNING, "RemoveNickFromChannel(): NULL user.");
    return;
  }

/* remove the user from the channel */
  ChannelUser = cptr->Joined;
  PrevUser = NULL;
  while (ChannelUser) 
  {
    if (!strcasecmp(ChannelUser->User->Nick, uptr->Nick)) 
    {
      TempCUser = ChannelUser;
      if (PrevUser)
	  PrevUser->Next = ChannelUser->Next;
      else
	  cptr->Joined = ChannelUser->Next;
      ChannelUser = ChannelUser->Next;
      free(TempCUser);
      ++DeletedChanUsers;
    }
    else 
    {
      PrevUser = ChannelUser;
      ChannelUser = ChannelUser->Next;
    }
  }

/* Remove the channel from the user */
  for(jptr = uptr->Chans;jptr;jptr = jptr->Next)
  {
       if(jptr->cptr == cptr)
       {
           if(jptr_prev)
               jptr_prev->Next = jptr->Next;
           else
               uptr->Chans = jptr->Next;
           /* TODO : When we're confident with this thing, these NULL things can be removed */
           jptr->cptr = NULL;
           free(jptr);
           jptr = NULL;
           break; /* Important because jptr->next is now invalid */
       }
       jptr_prev = jptr;
  }

  if (!DeletedChanUsers)
  {
    /* Because some kicks are followed by part, its quite possible for the user
     * to be removed twice, and this is acceptable. Dont wallop about it.
     */
    /* Debug(DBGWARNING, "RemoveNickFromChannel(): %s wasn't on %s!", uptr->Nick, cptr->Name); */
  }
  if (DeletedChanUsers > 1)
    Debug(DBGWARNING, "RemoveNickFromChannel(): Internal consistancy check error: %s was joined to %s multiple times. It is safe to continue.", uptr->Nick, cptr->Name);

  cptr->NumOfNicks -= DeletedChanUsers;

/** Done **/

  if(cptr->Regged == FALSE)
  {
     /* Channel is not registered */
     if(cptr->NumOfNicks < 1)
     {  /* If all the users have left   */
        /* then delete channel from memory list here.. */
        DelChan(cptr->Name);
     }
     return; /* Dont parse anything else for unreg'd channel. */
  }
  if(!cptr->settings[INCHAN])
    return;
  hptr = FindBestUserMatch(cptr, uptr);
  SetLastSeen(cptr, hptr, PART);
  if(hptr && (IsChanOwner(hptr) || IsChanCoOwner(hptr)))
  {
     cptr->OwnersOnline--;
     if(cptr->OwnersOnline < 0)
        cptr->OwnersOnline = 0;
     if(cptr->OwnersOnline == 0)
        cptr->LastAccess = time(NULL);
     WriteChanOpFile(cptr);
  }           
}

void ProcessPart(ChannelListType ChanPtr, UserListType UserNode)
{
  RemoveNickFromChannel(ChanPtr, UserNode);
}

void RemoveNickFromAllChannels(UserListType uptr)
{
  JoinedChansListType jptr, jptr_next;

  if(uptr == NULL)
     return;
  for(jptr = uptr->Chans; jptr; jptr = jptr_next)
  {
     jptr_next = jptr->Next;
     RemoveNickFromChannel(jptr->cptr, uptr);
     /* jptr is removed in this call */
  }
}

/* called when x2 is disconnected - reset the channel structures to something
 * reflecting the fact that they are empty */
void ClearAllChans()
{
   int i;
   ChannelListType cptr;
   BanListType bptr, nextbptr;

   for (i = 0; i < CHANSEGSIZE; i++)
  {
    cptr = Channels->Table[i];

    while(cptr)
    {
       cptr->OwnersOnline = 0;
       strcpy(cptr->CurrentKey, "");
       for(bptr = cptr->banned;bptr;bptr = nextbptr)
       {
           nextbptr = bptr->Next;
           free(bptr);
       }     
       cptr->banned = NULL;
       cptr->NumOfNicks = 0;
       if(cptr->Joined != NULL)
       {
          Debug(DBGWARNING, "We are clearing a channel that still has people joined to it.");
       }
       cptr->Joined = NULL; /* A memory leak, if the above was triggered.. but shouldnt happen. */
       cptr = cptr->Next;
    }
  }
}


int DelChan(char *Channel)
{
  ChannelListType CurrPtr, TailPtr;
  HandleListType hptr, hptr_next;
  ChannelUsersListType jptr, jptr_next;
  LamerListType lptr, lptr_next;
  BanListType bptr, Nextbptr;

  int Pos;
  long unsigned Key;

  TailPtr = NULL;

  Key = Hash(&Channel[1]);
  Pos = Key % CHANSEGSIZE;

  CurrPtr = Channels->Table[Pos];
  while (CurrPtr)
  {
    if (!strcasecmp(CurrPtr->Name, Channel))
    {
      break;
    }
    TailPtr = CurrPtr;
    CurrPtr = CurrPtr->Next;
  }
  if (!CurrPtr)
    return (0);
  if (!TailPtr)
    Channels->Table[Pos] = CurrPtr->Next;
  else
    TailPtr->Next = CurrPtr->Next;
/* Delete handle's & masks */
  for(hptr = CurrPtr->Handles;hptr;hptr = hptr_next)
  {
     hptr_next = hptr->Next;
     DelHandle(CurrPtr, hptr);
  }
/* Delete channeluser's entries */
  for(jptr = CurrPtr->Joined;jptr;jptr = jptr_next)
  {
      jptr_next = jptr->Next;
      free(jptr);
  }
  for(bptr = CurrPtr->banned;bptr;bptr = Nextbptr)
  {
       Nextbptr = bptr->Next;
       free(bptr);
  }
  CurrPtr->banned = NULL;
/* Delete lamer list */
  for(lptr = CurrPtr->Lamers;lptr;lptr = lptr_next)
  {
	  lptr_next = lptr->Next;
	  free(lptr);
  }

  if(CurrPtr->Topic)
     free(CurrPtr->Topic);
  if(CurrPtr->URL)
     free(CurrPtr->URL);
  if(CurrPtr->Description)
     free(CurrPtr->Description);
  if(CurrPtr->Reggedby)
     free(CurrPtr->Reggedby);
  if (CurrPtr->Key)
    free(CurrPtr->Key);
  if (CurrPtr->Limit)
    free(CurrPtr->Limit);
  free(CurrPtr->Name);
  free(CurrPtr);
  ChannelCount--;
  return(1);
}

void convertToUnreg(ChannelListType cptr)
{
  if(cptr->Regged == FALSE)
  {
    Debug(DBGWARNING, "Tried to unregister an unregistered channel");
    return;
  }
  if(cptr->Topic)
  {
    free(cptr->Topic);
    cptr->Topic = NULL;
  }
  if(cptr->Email)
  {
    free(cptr->Email);
    cptr->Email = NULL;
  }
  if(cptr->Description)
  {
    free(cptr->Description);
    cptr->Description = NULL;
  }
  if(cptr->URL)
  {
    free(cptr->URL);
    cptr->URL = NULL;
  }
  if(cptr->Reggedby)
  {
    free(cptr->Reggedby);
    cptr->Reggedby = NULL;
  }
  /* Need to clear userlists here, I guess. */
}

void convertToReg(ChannelListType cptr, char *Topic, char *Email, char *Description, char *URL, char *ReggedBy, char *ModeStr)
{
   int i;

   if(cptr->Regged)
   {
     Debug(DBGWARNING, "Tried to register an already registered channel");
     return;
   }
   cptr->Topic = malloc(strlen(Topic)+1);
   strcpy(cptr->Topic, Topic);

   cptr->Email = malloc(strlen(Email)+1);
   strcpy(cptr->Email, Email);

   cptr->Description = malloc(strlen(Description)+1);
   strcpy(cptr->Description, Description);

   cptr->URL = malloc(strlen(URL)+1);
   strcpy(cptr->URL, URL);

   cptr->Reggedby = malloc(strlen(ReggedBy)+1);
   strcpy(cptr->Reggedby, ReggedBy);

   cptr->Regged = TRUE;
   cptr->NumOfOps = 0;
   cptr->OwnersOnline = 0;
   
   cptr->NumOfNicks = 0;

   ConvertModeString(ModeStr, &cptr->PlusModes, &cptr->MinModes, &cptr->Key, &cptr->Limit);

   /* configure default channel settings */
   for(i=0;i<NUMOFCHANSETTINGS;i++)
       cptr->settings[i] = set_commands[i].def;

   /* Anything else?? */
}

/* Called with full params when reading data file; Called with mostly NULL params
 * when a user joins an unregistered channel.
 */
int AddChan(char *ChanName, int IsRegged, char *ModeString, char *ChanTopic, char *ChanEmail, char *Description, char *URL, char *ReggedBy, int settings[NUMOFCHANSETTINGS], time_t CreationDate, time_t LastAccess)
{
  ChannelListType NewChan;
  unsigned long Key;
  int Pos, i;
  if(GetChannel(ChanName))
  {
     Debug(DBGWARNING, "AddChan: Channel already in memory!");
     return(FALSE);
  }

  NewChan = (ChannelListType) malloc(sizeof(struct ChanStruct));

  if (!NewChan)
  {
    printf("Could not malloc ChanStruct! Terminating AddChan\n");
    return (0);
  }

  NewChan->Name = (char *) malloc(strlen(ChanName) + 1);
  if (!NewChan->Name)
  {
    Debug(DBGWARNING, "Could not malloc Name! Terminating AddChan\n");
    return (0);
  }
  if(ChanTopic)
  {
    NewChan->Topic = (char *) malloc(strlen(ChanTopic) + 1);
    if (!NewChan->Topic)
    {
      Debug(DBGWARNING, "Could not malloc Topic!  Terminating AddChan\n");
      return (0);
    }
  }
  if(ChanEmail)
  {
     NewChan->Email = (char *) malloc(strlen(ChanEmail) + 1);
     if(!NewChan->Topic)
     {
       Debug(DBGWARNING, "Could not malloc Email! Terminating AddChan\n");
       return (0);
     }
  }
  if(URL)
  {
     NewChan->URL = (char *) malloc(strlen(URL) + 1);
     if(!NewChan->URL)
     {
       printf("Could not malloc URL! Terminating AddChan\n");
       return(0);
     }
  }
  if(Description)
  {
    NewChan->Description = (char *) malloc(strlen(Description) + 1);
    if(!NewChan->Description)
    {
      printf("Could not malloc Description! Terminating AddChan\n");
      return(0);
    }
  }
  if(ReggedBy)
  {
     NewChan->Reggedby = (char *) malloc(strlen(ReggedBy) + 1);                               
     if(!NewChan->Reggedby)
     {                                                                                                 
        printf("Could not malloc Reggedby! Terminating AddChan\n");                               
        return(0);
     }
  }
  strcpy(NewChan->Name, ChanName);
  if(ChanTopic)
     strcpy(NewChan->Topic, ChanTopic);
  else
     NewChan->Topic = NULL;
  if(ChanEmail)
     strcpy(NewChan->Email, ChanEmail);
  else
     NewChan->Email = NULL;
  if(URL)
     strcpy(NewChan->URL, URL);
  else
     NewChan->URL = NULL;
  if(Description)
    strcpy(NewChan->Description, Description);
  else
    NewChan->Description = NULL;
  if(ReggedBy)
    strcpy(NewChan->Reggedby, ReggedBy);
  else
    NewChan->Reggedby = NULL;
  NewChan->Key = NULL;
  NewChan->Limit = NULL;
  NewChan->Regged = IsRegged;
  NewChan->Suspended = NULL;
  ConvertModeString(ModeString, &NewChan->PlusModes, &NewChan->MinModes, &NewChan->Key, &NewChan->Limit);
  if(settings)
  {
    for(i=0;i<NUMOFCHANSETTINGS;i++)
      NewChan->settings[i] = settings[i];
  }
  else
  {
    for(i=0;i<NUMOFCHANSETTINGS;i++)
      NewChan->settings[i] = set_commands[i].def;
  }
  if(IsRegged == FALSE)
     NewChan->settings[INCHAN] = FALSE;
  NewChan->OwnersOnline = 0;
  NewChan->NumOfNicks = 0;
  NewChan->LastAccess = LastAccess;
  NewChan->CreationDate = CreationDate;
  NewChan->Handles = NULL;
  NewChan->Lamers = NULL;
  NewChan->Joined = NULL;
  NewChan->banned = NULL;

  ChannelCount++;

  Key = Hash(&ChanName[1]);
  Pos = Key % CHANSEGSIZE;

  NewChan->Next = Channels->Table[Pos];
  Channels->Table[Pos] = NewChan;
  return 1;
}

void WriteChanOpFile(ChannelListType ChanPtr)
{
  char sets_str[MAXLEN];
  int i;
  FILE *FilePtr;
  char str[MAXLEN], *UpCaseChan;
  char ModeString[MAXLEN];
  HandleListType hptr;
  MaskListType mptr;
  LamerListType lptr;

  if (!ChanPtr)
  {
      Debug(DBGSEVERE, "WriteChanOpFile() NULL channel");
      return;
  }
  if(ChanPtr->Regged == FALSE)
  {
     Debug(DBGWARNING, "Cannot save unregistered channel!");
     return;
  }
  FilePtr = fopen("./tmp/.tmpopfile", "w");
  if(!FilePtr)
  {
     Debug(DBGSEVERE, "*** ERROR OPENING TEMP OPFILE! :%s", strerror(errno));
     return;
  }
  clearerr(FilePtr);
/* Compile a string of setting values... */
  strcpy(sets_str, "");
  for(i=0;i<NUMOFCHANSETTINGS;i++)
  {
    char buff[MAXLEN];
    sprintf(buff, "%d", ChanPtr->settings[i]);
    if(i>0)
      strcat(sets_str, " ");
    strcat(sets_str, buff);
  }
  /* These MUST all start with different letters! the read fucntion is 
   * optomized to look only at (T) for topic, (M) for MOTD etc.
   * Commented lines are not yet implimented (but this format
   * is great coz when they are nothing will break!)
   */
  fprintf(FilePtr, "VERSION 4\n");
  fprintf(FilePtr, "CHANNEL %s\n", ChanPtr->Name);
  fprintf(FilePtr, "TOPIC %s\n", ChanPtr->Topic);
  fprintf(FilePtr, "EMAIL %s\n", ChanPtr->Email);
  fprintf(FilePtr, "DESCRIPTION %s\n", ChanPtr->Description);
  fprintf(FilePtr, "WEB %s\n", ChanPtr->URL);
  fprintf(FilePtr, "MODES %s\n", GetModeString(ModeString, ChanPtr->PlusModes, ChanPtr->MinModes, ChanPtr->Key, ChanPtr->Limit));
  fprintf(FilePtr, "SETTINGS %s\n", sets_str);
  fprintf(FilePtr, "LASTVISITED %lu\n", ChanPtr->LastAccess);
  fprintf(FilePtr, "REGEDBY %s %lu\n", ChanPtr->Reggedby , ChanPtr->CreationDate);
  /* End of header section.. on to USER section */

  /* Suspended is optional and must thus come AFTER ReggedBy */
  if (ChanPtr->Suspended)
	  fprintf(FilePtr, "PSUSPENDED %s\n", ChanPtr->Suspended);

  for(hptr = ChanPtr->Handles;hptr;hptr = hptr->Next)
  {
      fprintf(FilePtr, "HANDLE %s %c %d %lu %s %s\n",
                    hptr->Handle,
                    hptr->Level,
                    hptr->AutoOp,
                    hptr->LastSeen,
                    hptr->Pass,
                    hptr->Info);
      for(mptr = hptr->Masks;mptr;mptr = mptr->Next)
      {
          fprintf(FilePtr, "K %s\n", mptr->Mask);
      }
  }
  for(lptr = ChanPtr->Lamers;lptr;lptr = lptr->Next)
  {
     fprintf(FilePtr, "BAN %lu %lu %s %s\n",
                        lptr->Time,
                        lptr->LastSeen,
                        lptr->Mask,
                        lptr->Reason);
  }
  if(ferror(FilePtr))
  {
     Debug(DBGSEVERE, "Error writing to a tempopfile (Out of disk space?). %s data NOT saved.", ChanPtr->Name);
     fclose(FilePtr);
     sleep(4);
     exit(2);
  }
  fclose(FilePtr);
  UpCaseChan = UpCase(ChanPtr->Name);
  sprintf(str, "./CHANS/%s", &UpCaseChan[1]);
  unlink(str);
  rename("./tmp/.tmpopfile", str);
  free(UpCaseChan);
}

char *ConvertChanMode(char *StrModes, unsigned char Modes)
{

  unsigned char Mode;
  char *tmp;

  tmp = StrModes;
  Mode = CMODEn;
  if ((Mode & Modes) != 0)
    *StrModes++ = 'n';

  Mode = CMODEt;
  if ((Mode & Modes) != 0)
    *StrModes++ = 't';

  Mode = CMODEs;
  if ((Mode & Modes) != 0)
    *StrModes++ = 's';

  Mode = CMODEp;
  if ((Mode & Modes) != 0)
    *StrModes++ = 'p';

  Mode = CMODEi;
  if ((Mode & Modes) != 0)
    *StrModes++ = 'i';

  Mode = CMODEm;
  if ((Mode & Modes) != 0)
    *StrModes++ = 'm';

  Mode = CMODEk;
  if ((Mode & Modes) != 0)
    *StrModes++ = 'k';

  Mode = CMODEl;
  if ((Mode & Modes) != 0)
    *StrModes++ = 'l';

  *StrModes = '\0';
  return (tmp);
}

char *GetModeString(char *ModeString, int PlusModes, int MinModes, char *key, char *limit)
{
  char tmpstr[LINEBUFFSIZE];

  strcpy(ModeString, "");
  if (PlusModes)
  {
    strcat(ModeString, "+");
    strcat(ModeString, ConvertChanMode(tmpstr, PlusModes));
  }
  else
  {  /* Always put the + */
      strcat(ModeString, "+");
  }
  if (MinModes)
  {
    strcat(ModeString, "-");
    strcat(ModeString, ConvertChanMode(tmpstr, MinModes));
  }
  if (PlusModes & CMODEk)
  {
    strcat(ModeString, " ");
    strcat(ModeString, key);
  }
  if (PlusModes & CMODEl)
  {
    strcat(ModeString, " ");
    strcat(ModeString, limit);
  }
  if(!*ModeString) /* Empty mode is bad in the data files.. put a + in there */
     strcat(ModeString, "+");
  return (ModeString);
}

void add_ban(ChannelListType cptr, char *banstring)
{
  BanListType New_ban;

  if ((New_ban = (BanListType) malloc(sizeof(*cptr->banned))) == NULL)
    return;
  if (search_ban(cptr, banstring) != NULL)
    return;
  strncpy(New_ban->banstring, banstring, 69);
  New_ban->banstring[69] = '\0';
  New_ban->time = time(NULL);
  Debug(DBGINFO, "*** Adding a ban to the list.  banstring=%s\n", banstring);
  New_ban->Next = cptr->banned;
  cptr->banned = New_ban;
}

int delete_ban(ChannelListType cptr, char *banstring)
{
  BanListType bptr;
  BanListType theban;
  BanListType PrevBan;

  if ((theban = search_ban(cptr, banstring)) == NULL)
    return (FALSE);

  for (bptr = PrevBan = cptr->banned; bptr; bptr = bptr->Next)
  {
    if (bptr == theban)
    {
      if(bptr == cptr->banned)
         cptr->banned = bptr->Next;
      else
         PrevBan->Next = bptr->Next;
      free(bptr);
      return (TRUE);
    }
    PrevBan = bptr;
  }
  return (FALSE);
}

int DeleteExpiredBans(ChannelListType cptr, BanListType * b_list, time_t period)
{
  BanListType Dummy, tmp;
  char unbanmode[10];
  char unbanstring[1024];
  int i;
  int found = 0;
  time_t currtime;

  /*Debug(DBGNOTICE, "*** Checking for bans in %s older than %lu\n", cptr->Name, period); */
  Dummy = *b_list;
  currtime = time(NULL);
  while (Dummy)
  {
    /* We can't just -bbb ban1 ban2,
       because the third b will make the server
       show the banlist etc. */
    strcpy(unbanmode, "-");
    strcpy(unbanstring, "");
    i = 0;
    while (Dummy && (i < 6))
    {
      if ((currtime - Dummy->time) > period)
      {
	strcat(unbanmode, "b");
	strcat(unbanstring, " ");
	strcat(unbanstring, Dummy->banstring);
	i++;
	found++;
	tmp = Dummy->Next;
	delete_ban(cptr, Dummy->banstring);
	Dummy = tmp;
      }
      else
	Dummy = Dummy->Next;
    }
    if (unbanmode[1])
      send_chanmode(cptr, "%s %s", unbanmode, unbanstring);
  }
  if (found > 0)
    return (1);
  else
    return (0);
}

void SaveDel(ChannelListType cptr)
{
  FILE *ChanFile;
  FILE *NewChanFile;
  char *UpCaseChan;
  char str[LINEBUFFSIZE];
  char ChanName[LINEBUFFSIZE];
  int s, line = 0;

  ChanFile = fopen(CHANFILENAME, "r");
  if (ChanFile == NULL)
  {
    Debug(DBGSEVERE, "Error: op file doesn't not exist!");
    return;
  }
  NewChanFile = fopen("tmp/.tmpchanfile.dat", "w");
  if(!NewChanFile)
  {  
     Debug(DBGSEVERE, "Error opening temp file in unreg()!!");
     return;
  }
  while ((s = fscanf(ChanFile, "%s\n", ChanName)) != EOF)
  {
    line++;
    if (s != 1)
    { 
      Debug(DBGERROR, "Error with param count in chan.dat! s=%d line=%d", s, line);
      fclose(ChanFile);
      fclose(NewChanFile);
      return;
    }
    if (strcasecmp(ChanName, cptr->Name))
      fprintf(NewChanFile, "%s\n", ChanName);
  }
  if(ferror(NewChanFile))
  {  
     Debug(DBGSEVERE, "Error writing to .tmpchanfile.dat (Out of disk space?). Unregistration NOT saved.", 
                           cptr->Name);
     fclose(NewChanFile);
     sleep(4);
  }
  fclose(ChanFile);
  fclose(NewChanFile);
  sprintf(str, "./%s", CHANFILENAME);
  unlink(str);
  rename("tmp/.tmpchanfile.dat", str);                                                        
  sprintf(str, "./CHANS/%s", UpCaseChan = UpCase(&cptr->Name[1]));                                   
  unlink(str);                                                                                
  free(UpCaseChan);                
}

void UnregExpired(ChannelListType cptr) 
{
  FILE *ChanFile;
  FILE *NewChanFile;
  char *UpCaseChan;
  char str[LINEBUFFSIZE];
  char ChanName[LINEBUFFSIZE];
  int s, line = 0;
  HandleListType tmphptr, nexthptr;
  LamerListType lptr, nextlptr;

  ChanFile = fopen(CHANFILENAME, "r");
  if (ChanFile == NULL)
  {
    Debug(DBGSEVERE, "\002Error:\002 Op file doesn't not exist!");
    return;
  }
  NewChanFile = fopen("tmp/.tmpchanfile.dat", "w");
  if(!NewChanFile)
  {
     Debug(DBGSEVERE, "Error opening temp file in unreg()!!");
     return;
  }
  while ((s = fscanf(ChanFile, "%s\n", ChanName)) != EOF)
  {
    line++;
    if (s != 1)
    {
      Debug(DBGERROR, "Error with param count in chan.dat! s=%d line=%d", s, line);
      fclose(ChanFile);
      fclose(NewChanFile);
      return;
    }
    if (strcasecmp(ChanName, cptr->Name))
      fprintf(NewChanFile, "%s\n", ChanName);
  }
  if(ferror(NewChanFile))
  {  
     Debug(DBGSEVERE, "\002Error:\002 Failure writing .tmpchanfile.dat (out of disk space?)!  Unregistration NOT saved.", cptr->Name);
     fclose(NewChanFile);
     sleep(4);
     exit(2);
  }
  fclose(ChanFile);
  fclose(NewChanFile);
  sprintf(str, "./%s", CHANFILENAME);
  unlink(str);
  rename("tmp/.tmpchanfile.dat", str);
  sprintf(str, "./CHANS/%s", UpCaseChan = UpCase(&cptr->Name[1]));
  unlink(str);
  free(UpCaseChan);
  send_part(cptr, "Channel expired");
/** If we were the only thing in the channel, delete it. **/
  if(cptr->NumOfNicks == 0)
    DelChan(cptr->Name);
  else
  {  /* Convert it back to an unreg'd channel.. */

     /* ModeString */
     if(cptr->Key)
     {
        free(cptr->Key);
        cptr->Key = NULL;
     }
     if(cptr->Limit)
     {
        free(cptr->Limit);
        cptr->Limit = NULL;
     }
     /* Topic */
     free(cptr->Topic);
     cptr->Topic = NULL;
     /* Email */
     free(cptr->Email);
     cptr->Email = NULL;
     /* Desc */
     free(cptr->Description);
     cptr->Description = NULL;
     /* URL... */
     free(cptr->URL);
     cptr->URL = NULL;
     /* ReggedBy... */
     free(cptr->Reggedby);
     cptr->Reggedby = NULL;

/* Delete userlist, and lamerlist.. */
     for(tmphptr = cptr->Handles;tmphptr;tmphptr = nexthptr)
     {
        nexthptr = tmphptr->Next;
        DelHandle(cptr, tmphptr);
     }
     cptr->Handles = NULL;
     for(lptr = cptr->Lamers;lptr;lptr = nextlptr)
     {
        nextlptr = lptr->Next;
        free(lptr);
     }
     cptr->Lamers = NULL;

     /* Regged...*/
     cptr->Regged = FALSE;
    /* The channel is now Unregistered. */
  }
  return;
}

void ExpireChannels(UserListType uptr)
{
  ChannelListType TempChanPtr, NextChanPtr;
  int i;
  FILE *logfile;
  time_t ct;
  char lvistime[LINEBUFFSIZE];
  char tmp[LINEBUFFSIZE];

  for (i = 0; i < CHANSEGSIZE; i++)
  {
    TempChanPtr = Channels->Table[i];

    while (TempChanPtr)
    {
      NextChanPtr = TempChanPtr->Next;
      if((TempChanPtr->Regged) && !(TempChanPtr->settings[NODELETE]) && (TempChanPtr->LastAccess > 0)) /* If its not marked nodelete */
      {
        if ((time(NULL) - TempChanPtr->LastAccess) > 86400 * CHANEXPIREDAYS) /* older than CHANEXPIREDAYS days */
        {
          Debug(DBGNOTICE, "Channel %s not visited since %lu. Removing..", TempChanPtr->Name,
                                                                             TempChanPtr->LastAccess);

          logfile = fopen(ExpiredChanLog, "a");
	  if (logfile != 0)
	  {
	    ct = time(NULL);
	    strcpy(tmp, ctime(&ct));
	    tmp[strlen(tmp) - 1] = '\0';
	    strcpy(lvistime, ctime(&TempChanPtr->LastAccess));
	    fprintf(logfile, ":%s: Channel %s removed, not visited since %s", tmp,
							TempChanPtr->Name, lvistime);
	    fclose(logfile);
	  }

          if(uptr)
             send_to_user(uptr, "Removing %s", TempChanPtr->Name);
	  UnregExpired(TempChanPtr);
        }
      }
      TempChanPtr = NextChanPtr;
    }
  }
}

void CheckBanTimeouts()
{
  int i;
  ChannelListType TempChanPtr;
  time_t period;

  Debug(DBGINFO, "*** Scanning for expired bans...\n");
  for (i = 0; i < CHANSEGSIZE; i++)
  {
    TempChanPtr = Channels->Table[i];
    while (TempChanPtr)
    {
      switch (TempChanPtr->settings[BANTIMEOUT])
      {
      case 0:
	period = -1;
	break;
      case 1:
	period = 600;		/* 10 mins */
	break;
      case 2:
	period = 7200;	/* 2 hours */
	break;
      case 3:
	period = 86400;	/* 24 hours */
	break;
      case 4:
	period = 604800;	/* one week */
	break;
      default:
	period = -1;
	break;
      }
      if (period > 0)
        DeleteExpiredBans(TempChanPtr, &(TempChanPtr->banned), period);
      TempChanPtr = TempChanPtr->Next;
    }
  }
}

int channel_massunban(ChannelListType Channel)
{
  BanListType bptr, oldbptr;
  char unbanmode[10];
  char unbanstring[1024];
  int i;
  int found = 0;

  if (Channel == NULL)
  {
    return(0); /* Channel not found */
  }
  if(Channel->settings[INCHAN] == FALSE)
     return(0); /* Not in channel */
  bptr = Channel->banned;
  while (bptr)
  {
    /* We can't just -bbb ban1 ban2,
       because the third b will make the server
       show the banlist etc. */
    strcpy(unbanmode, "-");
    strcpy(unbanstring, "");
    i = 0;
    while (bptr && (i < 6))
    {
      strcat(unbanmode, "b");
      strcat(unbanstring, " ");
      strcat(unbanstring, bptr->banstring);
      i++;
      oldbptr = bptr;
      bptr = bptr->Next;
      free(oldbptr);
      found++;
    }
    Channel->banned = NULL;
    send_chanmode(Channel, "%s %s", unbanmode, unbanstring);
  }
  if (found > 0)
    return (1);
  else
    return (0);
}

int channel_unban(ChannelListType Channel, char* banstring)
{
  BanListType bptr, tmp;
  char unbanmode[5];
  char unbanstring[MAXLEN];
  int i;
  int found = 0;

  if (Channel == NULL)
    return(0);
  if(Channel->settings[INCHAN] == FALSE)
  {
    return(0); /* Not in channel */
  }
  bptr = Channel->banned;
  while (bptr)
  {
    strcpy(unbanmode, "-");
    strcpy(unbanstring, "");
    i = 0;
    while (bptr && (i < 3))
    {
      tmp = bptr->Next;
      if (!matches(bptr->banstring, banstring) || !strcmp(banstring, bptr->banstring))
      {
	strcat(unbanmode, "b");
	strcat(unbanstring, " ");
	strcat(unbanstring, bptr->banstring);
	Debug(DBGINFO, "*** channel_unban calling delete_ban for %s", bptr->banstring);
	delete_ban(Channel, bptr->banstring);
	i++;
	found++;
      }
      bptr = tmp;
    }
    send_chanmode(Channel, "%s %s", unbanmode, unbanstring);
  }
  if (found > 0)
    return (1);
  else
    return (0);
}

BanListType search_ban(ChannelListType cptr, char *banstring)
{
  BanListType bptr;

  for (bptr = cptr->banned; bptr; bptr = bptr->Next)
    if (!strcasecmp(bptr->banstring, banstring))
      return (bptr);
  return (NULL);
}


/* Allocate a new mask struct, and throw it on a linked list. */
MaskListType AddMask(HandleListType hptr, char *Mask)
{
   MaskListType NewMask;

   NewMask = (MaskListType) malloc(sizeof(struct MaskStruct));

   NewMask->Mask = (char *) malloc(strlen(Mask) + 1);

   strcpy(NewMask->Mask, Mask);

   /* Add to linked list.. */
   NewMask->Next = hptr->Masks;
   hptr->Masks = NewMask;

   return(NewMask);
}

/* Allocate a new handle struct and throw it on a linked list. */
HandleListType AddHandle(ChannelListType cptr, char *Handle, char *Pass, char *Info, char Level, int AutoOp, time_t LastSeen, int InChan)
{
  HandleListType NewHandle;

  NewHandle = (HandleListType) malloc(sizeof(struct HandleStruct));

  NewHandle->Handle = (char *) malloc(strlen(Handle) + 1);
  NewHandle->Pass   = (char *) malloc(strlen(Pass) + 1);
  NewHandle->Info   = (char *) malloc(strlen(Info) + 1);

  strcpy(NewHandle->Handle, Handle);
  strcpy(NewHandle->Pass, Pass);
  strcpy(NewHandle->Info, Info);

  NewHandle->Level = Level;
  NewHandle->AutoOp = AutoOp;
  NewHandle->LastSeen = LastSeen;
  NewHandle->InChan = InChan;
  NewHandle->AuthUsers = NULL;

  NewHandle->Masks = NULL;  /* No masks yet */

  /* Now add to the linked list */
  NewHandle->Next = cptr->Handles;
  cptr->Handles = NewHandle;

  /* update netstats */
  switch (Level)
  {
	case 'P': ++PeonCount; break;
	case 'O': ++OpCount; break;
	case 'M': ++ManagerCount; break;
	case 'C': ++CoOwnerCount; break;
	case 'W': ++OwnerCount; break;
  }
  /* and return a ptr to our new handle */
  return(NewHandle);
}

HandleListType SetHandle(HandleListType Handle, char *NewHandle)
{
    free(Handle->Handle);
    Handle->Handle = malloc(strlen(NewHandle) + 1);
    strcpy(Handle->Handle, NewHandle);
    return(Handle);
}

HandleListType FindHandle(ChannelListType cptr, char* Handle)
{
   HandleListType hptr;

   for(hptr = cptr->Handles;hptr;hptr=hptr->Next)
   {
      if(!strcasecmp(hptr->Handle, Handle))
      {
          return(hptr);
      }
   }
   return(NULL);
}

MaskListType FindMask(HandleListType hptr, char* Mask)
{
   MaskListType mptr;

   for(mptr = hptr->Masks;mptr;mptr=mptr->Next)
   {
      if(!strcasecmp(mptr->Mask, Mask))
        return(mptr);
   }
   return(NULL);
}

/* Returns True if the mask was found & removed. 
 * False if it wasnt found.
 */
int DelMask(HandleListType hptr, char* Mask)
{
    MaskListType tmpMaskPtr;
    MaskListType prevMaskPtr = NULL;

    for(tmpMaskPtr = hptr->Masks;tmpMaskPtr;tmpMaskPtr = tmpMaskPtr->Next)
    {
       if(!strcasecmp(Mask, tmpMaskPtr->Mask))
       {
          /* Redirect the linked list.. */
          if(prevMaskPtr)
          {
             /* point the previous record to the next */
             prevMaskPtr->Next = tmpMaskPtr->Next;
          }
          else
          {
              /* were first; point hptr->Masks to the next */
              hptr->Masks = tmpMaskPtr->Next;
          }
          /* Free the memory */
          if(tmpMaskPtr->Mask != NULL)
          {
             free(tmpMaskPtr->Mask);
             tmpMaskPtr->Mask = NULL;
          }
          free(tmpMaskPtr); 
          tmpMaskPtr = NULL;
          /* its gone.. & even initalized.*/
          return(TRUE);
       }
       prevMaskPtr = tmpMaskPtr;
    }
    /* Not found */
    return(FALSE);
}

/* Delete all the masks for the handle
 * and then remove it.
 */
void DelHandle(ChannelListType cptr, HandleListType hptr)
{
    HandleListType tmpHandlePtr;
    HandleListType prevHandlePtr=NULL;
    AuthUserListType auptr, auptr_next;

    for(tmpHandlePtr = cptr->Handles;tmpHandlePtr;tmpHandlePtr = tmpHandlePtr->Next)
    {
        if(tmpHandlePtr == hptr)
        {
          /* First remove all the masks */
           MaskListType mptr = tmpHandlePtr->Masks;
           MaskListType nextmptr;

           while(mptr)          
           {
              nextmptr = mptr->Next; 
              /* Do the deleting of this one. dont worry aobut the list, since there ALL going away. */
              if(mptr->Mask)
                 free(mptr->Mask);
              free(mptr);
              /* Done deleting. */
              mptr = nextmptr;
           }
           /* now deal with the users (if any) logged in as this person */
           for(auptr = hptr->AuthUsers; auptr; auptr = auptr_next)
           {
               /* For every pointer we have to a user, delete that users's
                * pointer to us. */
               DelAuthHandlePtr(hptr, auptr->uptr);
               /* Now, free the memory, being carefull to preserve the Next
                * pointer for continuing the loop.. */
               auptr_next = auptr->Next;
               free(auptr);
           }
           /* now redirect the handle's linked list. */
           if(prevHandlePtr == NULL)
           {
                cptr->Handles = tmpHandlePtr->Next;
           }
           else
           {
                prevHandlePtr->Next = tmpHandlePtr->Next;
           }
           /* update netstats */
           switch (tmpHandlePtr->Level)
           {
	         case 'P': --PeonCount; break;
	         case 'O': --OpCount; break;
	         case 'M': --ManagerCount; break;
	         case 'C': --CoOwnerCount; break;
   	         case 'W': --OwnerCount; break;
           }
           /* Now free everything */
           if(tmpHandlePtr->Handle)
              free(tmpHandlePtr->Handle);
           if(tmpHandlePtr->Pass)
              free(tmpHandlePtr->Pass);
           if(tmpHandlePtr->Info)
              free(tmpHandlePtr->Info);
           free(tmpHandlePtr);
           /* were done. quit. */
           return;
        } /* if */
        prevHandlePtr = tmpHandlePtr;
    } /* for */
    /* Hmm none found. shouldnt happen.. */
    return;
} /* DelHandle */

int GetChanStats(ChannelListType cptr, int *w, int *c, int *m, int *o, int *p, int *l)
{
  HandleListType Handles;
  LamerListType Lamers;

  *w = *m = *o = *l = *p = *c = 0;
  Handles = cptr->Handles;
  while (Handles)
    {
      if (Handles->Level == 'W')
	++*w;
      else if (Handles->Level == 'C')
        ++*c;
      else if (Handles->Level == 'M')
	++*m;
      else if (Handles->Level == 'O')
	++*o;
      else if (Handles->Level == 'P')
        ++*p;
      else
      {
	Debug(DBGERROR, "I seem to have corruption in %s! Tell the admin GetChanStats found an unknown entry!!", cptr->Name);
        return(FALSE);
      }
    Handles = Handles->Next;
  }
  Lamers = cptr->Lamers;
  while (Lamers)
  {
      ++*l;
      Lamers = Lamers->Next;
  }
/* TODO : Add lamerlist counting code here for the b variable. */
  return(TRUE);
}



